import math
import random
import re
import sys
import time
from importlib.machinery import ModuleSpec
from importlib.util import find_spec, module_from_spec
from types import ModuleType
from typing import Optional, Union, List, Tuple

# @formatter:off
# noinspection SpellCheckingInspection
ul = "\u00a1-\uffff"  # Unicode letters range (must not be a raw string).

# IP patterns
ipv4_re = (
    r"(?:0|25[0-5]|2[0-4][0-9]|1[0-9]?[0-9]?|[1-9][0-9]?)"
    r"(?:\.(?:0|25[0-5]|2[0-4][0-9]|1[0-9]?[0-9]?|[1-9][0-9]?)){3}"
)
ipv6_re = r"\[[0-9a-f:.]+\]"  # (simple regex, validated later)

# Host patterns
hostname_re = r"[a-z" + ul + r"0-9](?:[a-z" + ul + r"0-9-]{0,61}[a-z" + ul + r"0-9])?"

# Max length for domain name labels is 63 characters per RFC 1034 sec. 3.1
domain_re = r"(?:\.(?!-)[a-z" + ul + r"0-9-]{1,63}(?<!-))*"
tld_re = (
    r"\."  # dot
    r"(?!-)"  # can't start with a dash
    r"(?:[a-z" + ul + "-]{2,63}"  # domain label
    r"|xn--[a-z0-9]{1,59})"  # or punycode label
    r"(?<!-)"  # can't end with a dash
    r"\.?"  # may have a trailing dot
)
host_re = "(" + hostname_re + domain_re + tld_re + "|localhost)"

url_re = (
    r"^(?:[a-z0-9.+-]*)://"  # scheme is validated separately
    r"(?:[^\s:@/]+(?::[^\s:@/]*)?@)?"  # user:pass authentication
    r"(?:" + ipv4_re + "|" + ipv6_re + "|" + host_re + ")"  # host
    r"(?::[0-9]{1,5})?"  # port
    r"(?:[/?#][^\s]*)?"  # resource path
    r"\Z"
)
# @formatter:on


def epoch_s() -> int:
    return math.floor(time.time_ns() / 1_000_000_000)


def epoch_ms() -> int:
    return math.floor(time.time_ns() / 1_000_000)


def ensure_module(name: str) -> Optional[ModuleType]:
    if name in sys.modules:
        return sys.modules.get(name, None)
    elif find_spec(name) is not None:
        spec: ModuleSpec = find_spec(name)
        module: ModuleType = module_from_spec(spec)
        sys.modules[name] = module
        spec.loader.exec_module(module)
        return module


def get_int_from_headers(
        headers: Union[dict, List[Tuple[str, str]]],
        key: str, default: Optional[int] = None
) -> Optional[int]:
    result = default
    try:
        if isinstance(headers, dict):
            headers = headers.items()
        candidates = [i for i in headers if len(i) > 1
                      and isinstance(i[0], str)
                      and (isinstance(i[1], int) or isinstance(i[1], float) or isinstance(i[1], str))
                      ]
        values = [i[1] for i in candidates if i[0].lower() == key.lower()]
        value = values[0] if len(values) > 0 else default
        result = int(value)
    finally:
        return result


def get_str_from_headers(
        headers: Union[dict, List[Tuple[str, str]]],
        key: str, default: Optional[str] = None
) -> Optional[str]:
    if isinstance(headers, dict):
        headers = headers.items()
    candidates = [i for i in headers if len(i) > 1 and isinstance(i[0], str) and isinstance(i[1], str)]
    values = [i[1] for i in candidates if i[0].lower() == key.lower()]
    value = values[0] if len(values) > 0 else default
    return value if isinstance(value, str) else default


def pseudo_normal_random(lower: int = 0, upper: int = 100, quality: int = 10) -> int:
    """Generate a value between min-max in a normal distribution"""
    values = sum([random.randint(lower, upper) for _ in range(quality)])
    return math.floor(values / quality)


def fake_request_start_ms(lower: int = 0, upper: int = 2_000) -> int:
    # Generate random request start to something up to a second ago for local testing
    return epoch_ms() - pseudo_normal_random(lower, upper)


def is_valid_url(url: str):
    return isinstance(url, str) and bool(re.match(url_re, url, re.IGNORECASE))



import json
import os
from enum import Enum
from typing import Dict

from dynoscale.constants import *
from dynoscale.utils import is_valid_url, ensure_module


def get_redis_urls_from_environ() -> Dict[str, str]:
    """Returns a dictionary of possible redis urls from all known redis_url environment variables"""
    redis_env_vars = (ENV_REDIS_URL, ENV_REDISTOGO_URL, ENV_OPENREDIS_URL, ENV_REDISGREEN_URL, ENV_REDISCLOUD_URL)
    return {name: os.environ[name] for name in redis_env_vars if name in os.environ}


class RunMode(Enum):
    PRODUCTION = 1
    DEVELOPMENT = 2


class Config:

    @property
    def is_valid(self) -> bool:
        if not (self.dyno and isinstance(self.dyno, str) and self.dyno.endswith('.1')):
            return False
        if not (self.url and isinstance(self.url, str) and is_valid_url(self.url)):
            return False
        return True

    @property
    def is_not_valid(self) -> bool:
        return not self.is_valid

    @property
    def run_mode_name(self):
        return RunMode.DEVELOPMENT.name if self.is_dev_mode else RunMode.PRODUCTION.name

    @property
    def is_rq_available(self) -> bool:
        return ensure_module('rq') is not None and ensure_module('redis') is not None and self.redis_urls

    def __init__(self):
        self.is_dev_mode = bool(os.environ.get(ENV_DEV_MODE, False))
        self.dyno = os.environ.get(ENV_HEROKU_DYNO)
        self.url = os.environ.get(ENV_DYNOSCALE_URL)
        self.redis_urls = get_redis_urls_from_environ()

    def __repr__(self) -> str:
        obj = {ENV_DEV_MODE: self.is_dev_mode,
               ENV_HEROKU_DYNO: self.dyno,
               ENV_DYNOSCALE_URL: self.url,
               'redis_urls': self.redis_urls
               }
        return json.dumps(obj, skipkeys=True, sort_keys=True)



import logging
from typing import Callable

from dynoscale.agent import DynoscaleAgent
from dynoscale.config import Config
from dynoscale.constants import HTTP_X_REQUEST_START
from dynoscale.utils import epoch_ms, get_int_from_headers, fake_request_start_ms


class DynoscaleWsgiApp:

    def __init__(self, app):
        self.logger: logging.Logger = logging.getLogger(f"{__name__}.{DynoscaleWsgiApp.__name__}")
        self.logger.debug("__init__")
        self.__app = app

        self.config = Config()
        if self.config.is_not_valid:
            self.logger.warning(f"{DynoscaleWsgiApp.__name__} will not "
                                f"initialize {DynoscaleAgent.__name__} with invalid config: {self.config}.")
            return  # User app will keep running, but queue times won't be logged
        self.ds_agent = DynoscaleAgent()
        self.logger.info(f"{DynoscaleWsgiApp.__name__} started in {self.config.run_mode_name} mode.")

    def __call__(self, environ: dict, start_response: Callable):
        # Under no circumstances should we ever stop user's app from receiving the request
        try:
            self.logger.debug("__call__")
            if self.config.is_valid:
                self.log_queue_time(environ)
        except Exception:
            self.logger.error("Unknown error, while processing Wsgi __call__")
        finally:
            return self.__app(environ, start_response)

    def log_queue_time(self, environ):
        # Under no circumstances should we ever stop user app from receiving the request
        try:
            self.logger.debug(f"log_queue_time (e:{environ})")
            log_start = epoch_ms()
            http_x_request_start = get_int_from_headers(
                environ,
                HTTP_X_REQUEST_START,
                fake_request_start_ms() if self.config.is_dev_mode else None
            )
            if http_x_request_start is not None:
                req_queue_time: int = log_start - http_x_request_start
                req_timestamp = int(log_start / 1_000)
                self.logger.debug(f"log_queue_time - Logging queue time {req_queue_time}")
                self.ds_agent.log_queue_time(req_timestamp, req_queue_time)
            else:
                self.logger.info("Can not calculate queue time.")
        except Exception as e:
            self.logger.error(f"Unknown error while attempting to log a queue time: {e}")



import csv
import logging
import os
import time
from dataclasses import dataclass
from io import StringIO
from json import JSONDecodeError
from typing import Optional, Union, Iterable, Callable

import requests
from requests import Session, PreparedRequest, Response, Request

from dynoscale import __version__
from dynoscale.config import Config
from dynoscale.repository import DynoscaleRepository, RecordOut

MAX_RECORD_AGE = 300.0  # Records older than this will be discarded before each upload
DEFAULT_PUBLISH_FREQUENCY = 30.0  # Minimum time between publishing records
KEY_LAST_PUBLISH_ATTEMPT = 'publish_last_attempt'
KEY_LAST_PUBLISH_SUCCESS = 'publish_last_success'
KEY_PUBLISH_FREQUENCY = 'publish_frequency'

logger = logging.getLogger(__name__)


@dataclass
class ConfigResponse:
    publish_frequency: float


def csv_from_records(records: Iterable[RecordOut]) -> bytes:
    """Generates a csv formatted string from Records"""
    buffer = StringIO()
    csv_writer = csv.writer(buffer)
    csv_writer.writerows([(r.timestamp, r.metric, r.source, r.metadata) for r in records])
    return buffer.getvalue().encode()


def dump_prepared_request(req: PreparedRequest, file=None):
    logger.debug("dump_prepared_request")
    print(
        '{}\r\n{}\r\n\r\n{}'.format(
            req.method + ' ' + req.url,
            '\r\n'.join('{}: {}'.format(k, v) for k, v in req.headers.items()),
            req.body.decode(),
        ),
        file=file
    )


def extract_config_response(response: Response) -> Optional[ConfigResponse]:
    logger.debug("extract_config_response")
    try:
        pf = response.json().get('config', {}).get("publish_frequency")
        return ConfigResponse(publish_frequency=float(pf)) \
            if pf and not isinstance(pf, bool) and float(pf) >= 0 \
            else None
    except JSONDecodeError as e:
        logger.error(f"extract_config_response - Error parsing response as JSON {e} : {response.text}")
    except AttributeError as e:
        logger.error(f"extract_config_response - Attribute error {e} : {response.text}")


def upload_payload(url: str, payload: bytes, dyno: str, session: Optional[Session] = None) -> Optional[Response]:
    response = None
    try:
        logger.debug("upload_payload")
        session = session if session else requests.Session()
        if not payload:
            logger.debug("upload_payload  - empty payload, exiting")
            return
        headers = {
            'Content-Type': 'text/csv',
            'User-Agent': f"dynoscale-python;{__version__}",
            'HTTP_X_DYNO': dyno
        }
        method = 'POST'
        prepared: PreparedRequest = session.prepare_request(
            Request(method=method, url=url, headers=headers, data=payload)
        )

        if logger.isEnabledFor(level=logging.DEBUG):
            dump_prepared_request(prepared)

        response = session.send(prepared)
        req_size = len(prepared.headers) + len(prepared.body)
        if response.ok:
            logger.info(f"Dynoscale successfully uploaded {req_size}bytes with status code {response.status_code}.")
        else:
            logger.warning(f"Dynoscale received status code {response.status_code} after uploading {req_size} bytes.")
    except Exception as e:
        logger.error(f"Dynoscale failed to upload data with and exception: {e}")
    finally:
        return response


class DynoscalePublisher:

    @property
    def last_publish_attempt(self) -> float:
        return self.repository.get(KEY_LAST_PUBLISH_ATTEMPT, 0)  # 0 here means never even though it really is 1970

    @last_publish_attempt.setter
    def last_publish_attempt(self, t: Optional[float]):
        self.repository[KEY_LAST_PUBLISH_ATTEMPT] = t

    @property
    def last_publish_success(self) -> Optional[float]:
        return self.repository.get(KEY_LAST_PUBLISH_SUCCESS, None)

    @last_publish_success.setter
    def last_publish_success(self, t: float):
        self.repository[KEY_LAST_PUBLISH_SUCCESS] = t

    @property
    def publish_frequency(self) -> float:
        freq = self.repository.get('publish_frequency')
        if not freq:
            freq = DEFAULT_PUBLISH_FREQUENCY
            self.publish_frequency = DEFAULT_PUBLISH_FREQUENCY
        return freq

    @publish_frequency.setter
    def publish_frequency(self, seconds: float):
        self.repository[KEY_PUBLISH_FREQUENCY] = seconds

    def __init__(self, repository_path: Optional[Union[str, bytes, os.PathLike]] = None):
        self.logger: logging.Logger = logging.getLogger(f"{__name__}.{DynoscalePublisher.__name__}")
        self.config: Config = Config()
        self.repository: DynoscaleRepository = DynoscaleRepository(repository_path)
        self.session: Session = Session()
        self.pre_publish_hook: Optional[Callable] = None

    def tick(self):
        # check if we should publish at all
        if self.should_publish():
            # if we have a valid pre_publish_hook call it safely
            if callable(self.pre_publish_hook):
                try:
                    self.pre_publish_hook()
                except Exception as e:
                    logger.error(f"DynoscalePublisher encountered error while executing pre_publish_hook: {e}")
            # and attempt to publish
            self.publish()

    def should_publish(self) -> bool:
        return self.last_publish_attempt + self.publish_frequency < time.time()

    def publish(self):
        # First prune the log of old records
        self.repository.delete_records_older_than(MAX_RECORD_AGE)
        # then check if the config is even valid and return if it isn't
        if self.config.is_not_valid:
            self.logger.warning("Can not publish with invalid config.")
            return
        # if config is valid, store the time we grabbed records from db
        start_time = time.time()
        # if it is valid, store the time of the last attempt publish, which is now
        self.last_publish_attempt = start_time
        records = self.repository.get_all_records()
        # if there are no records to be published exit
        if not records:
            self.logger.info("There is nothing to publish to Dynoscale.")
            return
        # otherwise, upload the records and
        self.logger.info(f"Will publish {len(records)} records to Dynoscale.")
        response = upload_payload(self.config.url, csv_from_records(records), self.config.dyno, self.session)
        # if the upload failed, exit
        if not (isinstance(response, Response) and response.ok):
            self.logger.warning("Error publishing to Dynoscale.")
            return
        self.logger.debug("publish - response ok")
        # since it was successful store the start_time
        self.last_publish_success = start_time
        # and delete all successfully uploaded logs
        self.repository.delete_records(records)
        # and attempt to grab config from the response
        config_response = extract_config_response(response)
        # if we received config response, and it has a new publish_frequency store it
        if config_response:
            self.logger.debug(f"publish - received config response: {config_response}")
            if config_response.publish_frequency != self.publish_frequency:
                self.publish_frequency = config_response.publish_frequency
                self.logger.info(
                    f"Dynoscale updated publish frequency, next publish in {config_response.publish_frequency}s.")



import logging
import math
import os
import time
from dataclasses import dataclass
from os.path import exists
from typing import Optional, Union, Tuple

from dynoscale.permadict import Permadict

DEFAULT_DYNOSCALE_REPOSITORY_FILENAME: str = 'dynoscale_repo.sqlite3'


@dataclass
class Record:
    timestamp: int
    metric: int
    source: str
    metadata: str


@dataclass
class RecordIn(Record):
    pass


@dataclass
class RecordOut(RecordIn):
    row_id: int


# noinspection SqlNoDataSourceInspection
# noinspection SqlResolve
class DynoscaleRepository(Permadict):
    """Storage for logs regarding the lifecycle of requests"""

    def __init__(self, path: Optional[Union[str, bytes, os.PathLike]] = None, **kwargs):
        self.logger: logging.Logger = logging.getLogger(f"{__name__}.{DynoscaleRepository.__name__}")
        self.logger.debug("__init__")
        db_filename = path if path else os.path.join(os.getcwd(), DEFAULT_DYNOSCALE_REPOSITORY_FILENAME)
        self.db_existed = exists(db_filename)
        super().__init__(db_filename)
        self._create_log_table()
        self.logger.info(f"Dynoscale {'opened' if self.db_existed else 'created'} repository {self.filename}.")

    def _create_log_table(self):
        self.logger.debug("_create_log_table")
        with self.cursor() as cur:
            cur.execute(
                'CREATE TABLE IF NOT EXISTS logs(timestamp INTEGER, metric INTEGER, source STRING, metadata STRING)'
            )

    def add_record(self, record: Record):
        self.logger.debug(f"add_record ({record.timestamp},{record.metric},{record.source},{record.metadata})")
        with self.cursor() as cur:
            cur.execute(
                'INSERT INTO logs (timestamp, metric, source, metadata) VALUES (?,?,?,?)',
                (record.timestamp, record.metric, record.source, record.metadata)
            )

    def get_all_records(self) -> Tuple[RecordOut]:
        self.logger.debug("get_all_records")
        with self.cursor() as cur:
            cur.execute("SELECT rowid, timestamp, metric, source, metadata FROM logs ORDER BY timestamp")
            rows = cur.fetchall()
            return tuple(RecordOut(row_id=r[0], timestamp=r[1], metric=r[2], source=r[3], metadata=r[4]) for r in rows)

    def delete_records(self, records: Tuple[RecordOut]):
        if not (records and isinstance(records, Tuple) and isinstance(records[0], RecordOut)):
            self.logger.debug(f"delete_records - Attempting to delete non-iterable: {records}")
            return
        self.logger.debug(f"delete_records - Deleting {len(records)} records.")
        row_id_tuples = [(r.row_id,) for r in records if hasattr(r, 'row_id')]
        with self.cursor() as cur:
            try:
                cur.executemany('DELETE FROM logs WHERE rowid = (?)', row_id_tuples)
            except Exception as e:
                self.logger.warning(f"DynoscaleRepository ran into an issue while deleting records {e}")
        self.__vacuum()

    def delete_records_before(self, t: float):
        self.logger.debug(f"delete_records_before {t}")
        with self.cursor() as cur:
            cur.execute('DELETE FROM logs WHERE timestamp < (?)', (math.ceil(t),))
        self.__vacuum()

    def delete_records_older_than(self, seconds: float):
        self.logger.debug(f"delete_records_older_than {seconds} seconds")
        self.delete_records_before(time.time() - seconds)

    def __vacuum(self):
        self.logger.debug("__vacuum")
        with self.cursor() as cur:
            cur.execute('VACUUM')



import logging
import os
import queue
import threading
from typing import Optional, Union

from dynoscale.config import Config
from dynoscale.publisher import DynoscalePublisher
from dynoscale.repository import Record

RECORD_SOURCE = "web"
RECORD_METADATA = ""

logger = logging.getLogger(__name__)
request_log_queue = queue.Queue()


def queue_time_logger(
        repository_path: Optional[Union[str, bytes, os.PathLike]] = None,
        enable_rq_logger: bool = False
):
    logger.debug("queue_time_logger")
    publisher = DynoscalePublisher(repository_path)
    if enable_rq_logger:
        from dynoscale.workers.rq_logger import DynoscaleRqLogger
        rq_logger = DynoscaleRqLogger(repository_path)
        publisher.pre_publish_hook = rq_logger.log_queue_times

    while True:
        record: Record = request_log_queue.get()
        logger.debug(f"queue_time_logger - got record from queue: {record}")
        publisher.repository.add_record(record)
        publisher.tick()


class DynoscaleAgent:

    def __init__(
            self,
            repository_path: Optional[Union[str, bytes, os.PathLike]] = None
    ):
        self.logger: logging.Logger = logging.getLogger(f"{__name__}.{DynoscaleAgent.__name__}")
        self.logger.debug("__init__")
        self.config = Config()
        if self.config.is_valid:
            self.t_logger = threading.Thread(
                target=queue_time_logger,
                daemon=True,
                kwargs={
                    'repository_path': repository_path,
                    'enable_rq_logger': self.config.is_rq_available,
                },
                name='Dynoscale'
            )
            self.t_logger.start()
            self.logger.info(f"Logging thread '{self.t_logger.name}' started.")
        else:
            self.logger.warning(
                f"{DynoscaleAgent.__name__} will not start Dynoscale logger thread with invalid config: {self.config}."
            )

    def log_queue_time(self, timestamp: int, queue_time: int):
        self.logger.debug(f"log_queue_time - {timestamp} {queue_time}")
        if self.config.is_valid:
            record = Record(timestamp, queue_time, RECORD_SOURCE, RECORD_METADATA)
            try:
                request_log_queue.put(record, timeout=0.25)
            except queue.Full:
                self.logger.error(f"Log queue is full, record {record} won't be logged!")
        else:
            self.logger.info(f"Throwing away queue time for dyno {self.config.dyno}.")



import datetime
import logging
import os
from contextlib import contextmanager
from typing import Optional, Union, Iterable, Generator

import redis
from rq import Queue
from rq.job import Job

from dynoscale.config import Config, get_redis_urls_from_environ
from dynoscale.repository import DynoscaleRepository, Record
from dynoscale.utils import epoch_s

logger = logging.getLogger(__name__)


def get_oldest_queued_job(queue) -> Optional[Job]:
    queued_jobs = [job for job in queue.jobs if job.is_queued]
    sorted_jobs = sorted(queued_jobs, key=lambda job: job.enqueued_at)
    return sorted_jobs[0] if sorted_jobs else None


def queue_time_records_for_connection(connection) -> Iterable[Record]:
    for queue in Queue.all(connection=connection):
        oldest_job = get_oldest_queued_job(queue)
        if oldest_job is not None:
            queue_time_ms = int((datetime.datetime.utcnow() - oldest_job.enqueued_at).total_seconds() * 1000)
            yield Record(timestamp=epoch_s(), metric=queue_time_ms, source=f"rq:{queue.name}", metadata="")


@contextmanager
def redis_connection(url: str) -> Generator[redis.Redis, None, None]:
    conn = None
    try:
        conn = redis.from_url(url)
        yield conn
    finally:
        if conn:
            conn.close()


def get_queue_time_records() -> Iterable[Record]:
    for url_name, redis_url in get_redis_urls_from_environ().items():
        try:
            with redis_connection(redis_url) as conn:
                for record in queue_time_records_for_connection(conn):
                    yield record
        except Exception as e:
            logger.warning(f"Dynoscale caught exception while getting queue times: {e}")


class DynoscaleRqLogger:
    def __init__(self, repository_path: Optional[Union[str, bytes, os.PathLike]] = None):
        self.logger: logging.Logger = logging.getLogger(f"{__name__}.{DynoscaleRqLogger.__name__}")
        self.config = Config()
        if not self.config.is_rq_available:
            self.logger.warning("Rq not available, DynoscaleRqLogger will not initialize.")
            return
        self.repository = DynoscaleRepository(repository_path)

    def log_queue_times(self):
        try:  # Under no circumstances should log_queue_times crash
            if not self.config.is_rq_available:
                self.logger.error("Rq not available, can't log RQ worker queue times.")
                return
            for record in get_queue_time_records():
                try:
                    self.repository.add_record(record)
                except Exception as e:
                    logger.warning(f"DynoscaleRqLogger couldn't add record: {record} exception: {e} ")
        except Exception as e:
            logger.warning(f"DynoscaleRqLogger couldn't get queue times: {e} ")



import pickle
import sqlite3
import sys
from contextlib import contextmanager

if sys.version_info[:2] >= (3, 8):
    from collections.abc import MutableMapping
else:
    from collections import MutableMapping


# noinspection SqlNoDataSourceInspection,SqlResolve
class Permadict(MutableMapping):
    """Persistent dict-like object backed by SQLite.

    :param filename: path to database or ``:memory:`` (the default)
    :param journal_mode: SQLite journal mode to use (default: "OFF")
    :param synchronous: when False (the default), set ``PRAGMA synchronous = OFF``
    :param kwargs: keyword arguments to initialize keys and values with

    """

    def __init__(self, filename=":memory:", journal_mode="OFF", synchronous=False, **kwargs):
        self.filename = filename
        self.conn = sqlite3.connect(self.filename)
        self._create_table(journal_mode, synchronous)

        if len(kwargs) > 0:
            for key, value in kwargs.items():
                self[key] = value

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        self.close()

    @contextmanager
    def cursor(self):
        with self.conn:
            cursor = self.conn.cursor()
            yield cursor
            cursor.close()

    def _create_table(self, journal_mode, synchronous):
        sql = [
            "CREATE TABLE IF NOT EXISTS dict (name BLOB PRIMARY KEY, object BLOB);",
            "CREATE INDEX IF NOT EXISTS ix_name ON dict (name);"
        ]

        if not synchronous:
            sql += ["PRAGMA synchronous = OFF;"]
        sql += ["PRAGMA journal_mode = {};".format(journal_mode)]

        with self.cursor() as cursor:
            for statement in sql:
                cursor.execute(statement)

    def __len__(self):
        with self.cursor() as cur:
            cur.execute("SELECT COUNT(*) FROM dict")
            return cur.fetchone()[0]

    def __getitem__(self, key):
        with self.cursor() as cur:
            cur.execute("SELECT object FROM dict WHERE name = (?)", (key,))
            obj = cur.fetchone()
            if obj is None:
                raise KeyError("No such key: " + key)
            key = obj[0]
            return pickle.loads(key)

    def __setitem__(self, key, value):
        with self.cursor() as cur:
            binary = sqlite3.Binary(pickle.dumps(value, protocol=pickle.HIGHEST_PROTOCOL))
            cur.execute("INSERT OR REPLACE INTO dict VALUES (?,?)", (key, binary))

    def __delitem__(self, key):
        if key in self:
            with self.cursor() as cur:
                cur.execute("DELETE FROM dict WHERE name = (?)", (key,))

    def __contains__(self, key):
        try:
            # noinspection PyStatementEffect
            self[key]
            return True
        except KeyError:
            return False

    def __iter__(self):
        for key in self.keys():
            yield key

    def keys(self):
        with self.cursor() as cur:
            cur.execute("SELECT name FROM dict")
            return [key[0] for key in cur.fetchall()]

    def items(self):
        for key in self:
            yield key, self[key]

    def values(self):
        """A generator which iterates over the :class:`Permadict`'s values."""
        for key in self:
            yield self[key]

    def clear(self):
        """Remove all items from the Permadict."""
        with self.cursor() as cur:
            # noinspection SqlWithoutWhere
            cur.execute("DELETE FROM dict")

    def get(self, key, default=None):
        """Return the value for ``key`` if it exists, otherwise return the
        ``default``.

        """
        try:
            return self[key]
        except KeyError:
            return default

    def pop(self, key):
        """If ``key`` is present, remove it and return its value, else raise a
        :class:`KeyError`.

        """
        try:
            value = self[key]
            del self[key]
            return value
        except KeyError:
            raise

    def update(self, iterable, **kwargs):
        """Update the :class:`Permadict` with the key/value pairs of
        ``iterable``.

        Returns ``None``.

        """
        if isinstance(iterable, dict):
            iter_ = iterable.items()
        else:
            iter_ = iterable
        for key, value in iter_:
            self[key] = value
        return None

    def close(self):
        self.conn.close()



import logging

from dynoscale.agent import DynoscaleAgent
from dynoscale.config import Config
from dynoscale.constants import X_REQUEST_START
from dynoscale.utils import epoch_ms, get_int_from_headers, fake_request_start_ms

PROC_NAME = '__dynoscale_hook_processor'

logger = logging.getLogger(__name__)


def pre_request(worker, req):
    logger.debug(f"pre_request - {worker} {req}")
    processor = globals().get(PROC_NAME, None)
    if processor is None:
        processor = GunicornHookProcessor()
        globals()[PROC_NAME] = processor
    processor.pre_request(worker, req)


class GunicornHookProcessor:

    def __init__(self):
        self.logger: logging.Logger = logging.getLogger(f"{__name__}.{GunicornHookProcessor.__name__}")
        self.config = Config()
        if self.config.is_valid:
            self.logger.info(f"Dynoscale started {GunicornHookProcessor.__name__} in {self.config.run_mode_name} mode.")
            self.ds_agent = DynoscaleAgent()
        else:
            self.logger.warning(
                f"Dynoscale can't start {GunicornHookProcessor.__name__} with invalid config: {self.config}."
            )
            return

    def pre_request(self, worker, req):
        # Under no circumstances should we ever stop user app from receiving the request
        try:
            self.logger.debug(f"pre_request (w:{id(worker)} w.pid{worker.pid} rq:{id(req)})")
            log_start: int = epoch_ms()
            http_x_request_start = get_int_from_headers(
                req.headers,
                X_REQUEST_START,
                fake_request_start_ms() if self.config.is_dev_mode else None
            )
            if http_x_request_start is not None:
                req_queue_time: int = log_start - http_x_request_start
                req_timestamp = int(log_start / 1_000)
                self.logger.debug(f"pre_request - Logging queue time: {req_queue_time}")
                self.ds_agent.log_queue_time(req_timestamp, req_queue_time)
            else:
                self.logger.info("Can not calculate queue time.")
        except Exception as e:
            self.logger.error(f"Unknown error while attempting to log a queue time: {e}")



# Environment Variables
ENV_DEV_MODE = 'DYNOSCALE_DEV_MODE'
ENV_HEROKU_DYNO = 'DYNO'
ENV_DYNOSCALE_URL = 'DYNOSCALE_URL'
# Redis
ENV_REDIS_URL = 'REDIS_URL'
ENV_REDISTOGO_URL = 'REDISTOGO_URL'
ENV_OPENREDIS_URL = 'OPENREDIS_URL'
ENV_REDISGREEN_URL = 'REDISGREEN_URL'
ENV_REDISCLOUD_URL = 'REDISCLOUD_URL'
# Header
HTTP_X_REQUEST_START = 'HTTP_X_REQUEST_START'
X_REQUEST_START = 'X-REQUEST-START'
